// Featurer.cpp: implementation of the Featurer class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "RangeImage.h"
#include "Featurer.h"
#include "math.h"
#include "Data.h"
#include "Geometry.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
int compare( const void *elem1, const void *elem2 );


Featurer::Featurer()
{
  m_statistic = FALSE;
  m_expo_aut = FALSE;
  m_thinning = FALSE;
}

Featurer::Featurer(BOOL stat_in)
{
  m_statistic = stat_in;
  m_expo_aut = FALSE;
}

Featurer::~Featurer()
{

}

void Featurer::StMarc(int iter, float param)
{
  int i;
  float Km, Kg, Ixx, Ixy, Iyy, Ix, Iy;

  int ROW = data->row;
  int COL = data->col;
  
  long npix = (long )COL*ROW;

  float (** N)[3] = (float (**)[3])new float*[ROW];
   for( i=0; i<ROW; i++){
		N[i] = new float[COL][3];		
	}

  float (** n)[3] = (float (**)[3])new float*[ROW];
   for( i=0; i<ROW; i++){
		n[i] = new float[COL][3];		
	}
     

  float **C = new float*[ROW];
   for(i=0; i<ROW; i++){
		C[i] = new float[COL];		
	}
   
  float **Kmax = new float*[ROW];
   for(i=0; i<ROW; i++){
		Kmax[i] = new float[COL];		
	}

  float **Kmin = new float*[ROW];
   for(i=0; i<ROW; i++){
		Kmin[i] = new float[COL];		
	}

  float **alpha_min = new float*[ROW];
   for(i=0; i<ROW; i++){
		alpha_min[i] = new float[COL];		
	}

  float **alpha_max = new float*[ROW];
   for(i=0; i<ROW; i++){
		alpha_max[i] = new float[COL];		
	}

  float beta_min, beta_max;
  int x, y;
  
  for ( y=0; y<ROW; y++){
    N[y][0][0] = 0;
    N[y][0][1] = 0;
    N[y][0][2] = -1;
	C[y][0] = 0;
	Kmin[y][0] = 0;
    Kmax[y][0] = 0;
  }

  for ( x=0; x<COL; x++)
  {	
    N[0][x][0] = 0;
	N[0][x][1] = 0;  
    N[0][x][2] = -1;
	C[0][x] = 0;
	Kmin[0][x] = 0;
	Kmax[0][x] = 0;
  }

  float **SIGMA = NULL;
  if (m_expo_aut){   
     SIGMA = new float*[ROW];
     for(i=0; i<ROW; i++){
		SIGMA[i] = new float[COL];		
	 }
  }

   for ( y=1; y<ROW-1; y++)
     for ( x=1; x< COL-1; x++)
	  {			
		Ix = 0.5f*(data->getValue(y, x+1) - data->getValue(y, x-1));
		Iy = 0.5f*(data->getValue(y+1, x) - data->getValue(y-1, x));
		float r = 1.0 + Ix*Ix + Iy*Iy;
		double sr = sqrt(r);

		N[y][x][0] = (float)(Ix/sr);
		N[y][x][1] = (float)(Iy/sr);
		N[y][x][2] = (float)(-1.0/sr);        

        float Ixx = -0.5* (N[y][x+1][0]/N[y][x+1][2] - N[y][x-1][0]/N[y][x-1][2] );     
		float Iyy = -0.5* (N[y+1][x][1]/N[y+1][x][2] - N[y-1][x][1]/N[y-1][x][2] );
		float Ixy = -0.25* (N[y+1][x][0]/N[y+1][x][2] - N[y-1][x][0]/N[y-1][x][2] 
			               + N[y][x+1][1]/N[y][x+1][2]- N[y][x-1][1]/N[y][x-1][2]);
			
		float Kg = (Ixx*Iyy - Ixy*Ixy)/(r*r);
		float Km = ((1.0 + Iy * Iy) * Ixx + (1.0 + Ix * Ix) * Iyy - 
			         2.0 * Ix * Iy * Ixy) / (2.0 * r * sr);
 
		float Delta = 2.0* Km * Km - Kg;  /* half-sum of the squared principal curvatures */

		if (m_expo_aut)
             C[y][x] = Delta;
		else
		     C[y][x] = exp(- param * Delta );
        
	  }

    // to fihd automatically exponential parameter
	//
	 if (m_expo_aut){   
       for ( y=1; y<ROW-1; y++)
        for ( x=1; x< COL-1; x++)
		{			
          float sigma = getSigma(y, x, N);
          float Delta = C[y][x];
          SIGMA[y][x] = sigma;
	      C[y][x] = exp(- sigma * Delta );
		}
	 }


  for (int it = 0; it < iter; it++)
      {
	  
 
  for (y = 1; y < ROW - 1; y++)
    for (int x = 1; x < COL - 1; x++)	
	  {
        float S=0;
        n[y][x][0] = 0;
        n[y][x][1] = 0;
        n[y][x][2] = 0;
	      
	for (i=-1; i <= 1; i++)	      
		for (int j=-1; j <= 1; j++) 
		  {
	          float dist = 1;
//		     if (i*i+j*j != 0){dist = sqrt(i*i+j*j+SQ(PIXEL(I,x+i,y+j) - PIXEL(I,x,y)));}
	     	  n[y][x][0] += N[y+j][x+i][0] * C[y+j][x+i] / dist;
              n[y][x][1] += N[y+j][x+i][1] * C[y+j][x+i] / dist;
              n[y][x][2] += N[y+j][x+i][2] * C[y+j][x+i] / dist;
              S += C[y+j][x+i] / dist;
	       }
            
         n[y][x][0] /= S;
         n[y][x][1] /= S;
         n[y][x][2] /= S;
	  }


	  for (y = 1; y <ROW - 1; y++) 
    	for ( x = 1; x < COL - 1; x++)                
          {
            Ix = - n[y][x][0] /n[y][x][2];
	        Iy = - n[y][x][1] /n[y][x][2];

	        float r  = 1 + Ix * Ix + Iy * Iy;
	        double sr = sqrt(r);

	       Ixx = - 0.5 * (n[y][x+1][0]/n[y][x+1][2] - n[y][x-1][0]/n[y][x-1][2]);
	       Iyy = - 0.5 * (n[y+1][x][1]/n[y+1][x][2] - n[y-1][x][1]/n[y-1][x][2]);
	       Ixy = - 0.25 * (n[y+1][x][0]/n[y+1][x][2] - n[y-1][x][0]/n[y-1][x][2] +
			               n[y][x+1][1]/n[y][x+1][2] - n[y][x-1][1]/n[y][x-1][2]);


	       Kg  = (Ixx * Iyy - Ixy * Ixy) / (r * r);
	       Km  = ((1 + Iy * Iy) * Ixx + (1 + Ix * Ix) * Iyy - 2 * Ix * Iy * Ixy) / (2. * r * sr);

           float Delta = 2.* Km * Km - Kg;

		   if (m_expo_aut )
		      C[y][x] = exp(- SIGMA[y][x] * Delta);
		   else
             C[y][x] = exp(- param * Delta);

           N[y][x][0] = n[y][x][0];
           N[y][x][1] = n[y][x][1];
           N[y][x][2] = n[y][x][2];
          }
       }
 
  if (SIGMA != NULL){
    for (int i=0; i<ROW; i++)
		delete SIGMA[i];
  
    delete [] SIGMA;
  }

  for (y=0;y < ROW; y++) 
	  delete [] n[y];
  delete [] n;

  float (** Vmax)[2] = (float (**)[2])new float*[ROW];
  for( i=0; i<ROW; i++)
		Vmax[i] = new float[COL][2];		
	

  float (** Vmin)[2] = (float (**)[2])new float*[ROW];
  for( i=0; i<ROW; i++)
		Vmin[i] = new float[COL][2];		

  for (y = 1; y < ROW - 1; y++)
    for ( x = 1; x < COL - 1; x++)    
	 {
        Ix = - N[y][x][0] /N[y][x][2];
	    Iy = - N[y][x][1] /N[y][x][2];

	    float r  = 1 + Ix * Ix + Iy * Iy;
	    double sr = sqrt(r);


	    Ixx = - 0.5 * (N[y][x+1][0]/N[y][x+1][2] - N[y][x-1][0]/N[y][x-1][2]);
	    Iyy = - 0.5 * (N[y+1][x][1]/N[y+1][x][2] - N[y-1][x][1]/N[y-1][x][2]);
	    Ixy = - 0.25 * (N[y+1][x][0]/N[y+1][x][2] - N[y-1][x][0]/N[y-1][x][2] +
			               N[y][x+1][1]/N[y][x+1][2] - N[y][x-1][1]/N[y][x-1][2]);


	     Kg  = (Ixx * Iyy - Ixy * Ixy) / (r * r);
	     Km  = ((1 + Iy * Iy) * Ixx + (1 + Ix * Ix) * Iyy - 2 * Ix * Iy * Ixy) / (2. * r * sr);

 	     float D = Km * Km - Kg;
	     if (D < 0.0) D = 0.0;

	     Kmax[y][x] = Km + sqrt(D);
	     Kmin[y][x] = Km - sqrt(D);

	     float Amax   = Ixx / sr - (1+ Ix * Ix) * Kmax[y][x];
	     float Bmax   = Ixy / sr - Ix * Iy * Kmax[y][x];
	     float Cmax   = Iyy / sr - (1  + Iy * Iy) * Kmax[y][x];

	     if (fabs(Amax) > fabs(Cmax)){
		   Vmax[y][x][0] = -Bmax/(float)sqrt(1+Ix*Ix);
		   Vmax[y][x][1] =  Amax/(float)sqrt(1+Iy*Iy);
	     }
	     else{
		   Vmax[y][x][0] =  Cmax/(float)sqrt(1+Ix*Ix);
		   Vmax[y][x][1] = -Bmax/(float)sqrt(1+Iy*Iy);
	     }

	     float Amin   = Ixx / sr - (1+ Ix * Ix) * Kmin[y][x];
	     float Bmin   = Ixy / sr - Ix * Iy * Kmin[y][x];
	     float Cmin   = Iyy / sr - (1  + Iy * Iy) * Kmin[y][x];

	     if (fabs(Amin) > fabs(Cmin)){
		  Vmin[y][x][0] = -Bmin;
		  Vmin[y][x][1] =  Amin;
	     }
	     else{
		  Vmin[y][x][0] =  Cmin;
		  Vmin[y][x][1] = -Bmin;
	     }
/*****/
    beta_max = Vmax[y][x][0]*Iy-Vmax[y][x][1]*Ix;
    beta_max /= sr;
    float det_max = sqrt(SQ(Vmax[y][x][0])+SQ(Vmax[y][x][1])+ 
                 SQ(Vmax[y][x][0]*Ix+Vmax[y][x][1]*Iy));
    if (det_max > 0.001)  beta_max /= det_max;
     else  beta_max = 0;
    alpha_max[y][x] = beta_max/sqrt(1-SQ(beta_max));

    beta_min = Vmin[y][x][0]*Iy-Vmin[y][x][1]*Ix;
    beta_min /= sr;
    float det_min = sqrt(SQ(Vmin[y][x][0])+SQ(Vmin[y][x][1])+ 
                 SQ(Vmin[y][x][0]*Ix+Vmin[y][x][1]*Iy));
    if (det_min > 0.001)  beta_min /= det_min;
     else  beta_min = 0;
    alpha_min[y][x] = beta_min/sqrt(1-SQ(beta_min));
/*****/

	  }
 
	 for (y=0;y < ROW; y++){ 
	  delete [] N[y];
	  delete [] C[y];
	 }

	 delete [] C;
	 delete [] N;

	 if (m_statistic == TRUE){ 
       trimm_Kmax(Kmax, 60, 30, m_hyster_max, m_curv_thresh_max);
       trimm_Kmin(Kmin, 60, 30, m_hyster_min, m_curv_thresh_min);
	 }

    non_max_suppr(Kmax, Kmin, Vmax, Vmin, alpha_max, fabs(m_curv_thresh_max));
    non_min_suppr(Kmin, Kmax, Vmin, Vmax, alpha_max, fabs(m_curv_thresh_min));

//    hysteresis(Kmax, RIDGE, fabs(m_hyster_max));
//    hysteresis(Kmin, RAVINE, fabs(m_hyster_min));

	if (m_thinning)
		thinning(RIDGE, Kmax);
	if (m_thinning)
		thinning(RAVINE, Kmin);


//	post_process();

	
	 for (y=0;y < ROW; y++){ 
	  delete [] Vmax[y];
	  delete [] Vmin[y];
	  delete [] Kmin[y];		
	  delete [] Kmax[y];
	  delete [] alpha_min[y];
      delete [] alpha_max[y];
	 }

	 delete [] Kmax;
	 delete [] Kmin;
	 delete [] Vmax;
	 delete [] Vmin;
	 delete [] alpha_min;
     delete [] alpha_max;
}

float eps = 0.00000001;
float THRESH = 0.000001;

void Featurer::non_min_suppr(float **Kmin, float **Kmax, float (**Vmin )[2], 
							 float (**Vmax )[2], float** alpha, float T)
{
  int x,y;
  float tg, lappr1, lappr2;
  float d, sq_dist1, sq_dist2, A; 

  int ROW = data->row;
  int COL = data->col;

  float **B = new float*[ROW];
  for(int i=0; i<ROW; i++){
		B[i] = new float[COL];	
		for (int j=0; j<COL; j++)
			B[i][j] =0.0;
	}
 
  for (y = 1; y < ROW - 1; y++)
    for (x = 1; x < COL - 1; x++)
    {	
	  if (fabs(Vmax[y][x][0]) < eps) {
	      lappr1 = Kmax[y+1][x]; lappr2 = Kmax[y-1][x];
		  d = 1;
	  }
	  else {
	      tg = Vmax[y][x][1] / Vmax[y][x][0];

	      if ((tg >= 0.) && (tg < 1.)) {
	         lappr1 = Kmin[y+1][x+1] * (1.- tg) + tg * Kmin[y][x+1];
	         lappr2 = Kmin[y-1][x-1] * (1.- tg) + tg * Kmin[y][x-1];
			 d = sqrt(1+tg*tg);
		  }
          else if ( tg >= 1.) {
	         lappr1 = (Kmin[y+1][x+1] * (tg - 1.) + Kmin[y+1][x]) / tg;
	         lappr2 = (Kmin[y-1][x-1] * (tg - 1.) + Kmin[y-1][x]) / tg;
             d = sqrt(1+1/tg*tg);
		  }
          else if ((tg <= 0.) && (tg >  -1.)) {
	         lappr1 = Kmin[y+1][x-1] * (1.+ tg) - tg * Kmin[y][x-1];
	         lappr2 = Kmin[y-1][x+1] * (1.+ tg) - tg * Kmin[y][x+1];
             d = sqrt(1+tg*tg);
		  }
	     else if ( tg <= -1.) {
	         lappr1 = (Kmin[y+1][x-1] * (tg + 1.) - Kmin[y+1][x]) / tg;
	         lappr2 = (Kmin[y-1][x+1] * (tg + 1.) - Kmin[y-1][x]) / tg;
             d = sqrt(1+1/tg*tg);
		 }
	  }

	  B[y][x] = (lappr1 - lappr2)/(2*d);
	  if(Vmin[y][x][0]*Vmin[y][x][1] - Vmax[y][x][0]*Vmax[y][x][1]< 0)
      {
        B[y][x] = -(lappr1 - lappr2)/(2*d);
      }  
	}


	for (y = 1; y < ROW - 1; y++)
     for (x = 1; x < COL - 1; x++)
     {	
		if ( Geometry::isZero(Vmin[y][x][0]) ){
	      lappr1 = Kmin[y+1][x]; lappr2 = Kmin[y-1][x];
          sq_dist1 = 1+SQ(data->getValue(y+1,x) - data->getValue(y,x));
          sq_dist2 = 1+SQ(data->getValue(y-1,x) - data->getValue(y,x)); 
	}
	else {
	  tg = Vmin[y][x][1] / Vmin[y][x][0];

	  if ((tg >= 0.) && (tg < 1.)) {
	    lappr1 = Kmin[y+1][x+1] * (1.- tg) + tg * Kmin[y][x+1];
	    lappr2 = Kmin[y-1][x-1] * (1.- tg) + tg * Kmin[y][x-1];
          sq_dist1 = 1+tg*tg + SQ(data->getValue(y+1,x+1) * (1.- tg) 
                     + tg * data->getValue(y,x+1) - data->getValue(y,x));
          sq_dist2 = 1+tg*tg + SQ(data->getValue(y-1,x-1) * (1.- tg) 
                     + tg * data->getValue(y,x-1) - data->getValue(y,x));
	  }
	  else if ( tg >= 1.) {
	    lappr1 = (Kmin[y+1][x+1] * (tg - 1.) + Kmin[y+1][x]) / tg;
	    lappr2 = (Kmin[y-1][x-1] * (tg - 1.) + Kmin[y-1][x]) / tg;
          sq_dist1 = 1+1/(tg*tg)+ SQ(data->getValue(y+1,x+1) * (tg - 1.)/tg
                             + data->getValue(y+1,x)/tg - data->getValue(y,x));
          sq_dist2 = 1+1/(tg*tg)+ SQ(data->getValue(y-1,x-1) * (tg - 1.)/tg
                              + data->getValue(y-1,x)/tg - data->getValue(y,x));
	  }

	  else if ((tg <= 0.) && (tg >  -1.)) {
 	    lappr1 = Kmin[y+1][x-1] * (1.+ tg) - tg *Kmin[y][x-1];
	    lappr2 = Kmin[y-1][x+1] * (1.+ tg) - tg * Kmin[y][x+1];
          sq_dist1 = 1+tg*tg + SQ(data->getValue(y+1,x-1) * (1.+ tg) 
                     - tg * data->getValue(y,x-1) - data->getValue(y,x));
          sq_dist2 = 1+tg*tg + SQ(data->getValue(y-1,x+1) * (1.+ tg) 
                      - tg * data->getValue(y,x+1) - data->getValue(y,x));
	  }
	  else if ( tg <= -1.) {
	    lappr1 = (Kmin[y+1][x-1] * (tg + 1.) - Kmin[y+1][x]) / tg;
	    lappr2 = (Kmin[y-1][x+1] * (tg + 1.) - Kmin[y-1][x]) / tg;
          sq_dist1 = 1+1/(tg*tg)+ SQ(data->getValue(y+1,x-1) * (tg + 1.)/tg
                            - data->getValue(y+1,x)/tg - data->getValue(y,x));
          sq_dist2 = 1+1/(tg*tg)+SQ(data->getValue(y-1,x+1) * (tg + 1.)/tg
                            - data->getValue(y-1,x)/tg - data->getValue(y,x));
	  }
	}
        
	A = 0;
    if ( (Kmax[y][x]-Kmin[y][x]) > 0.0001 )
         A = SQ(B[y][x])/(Kmax[y][x]-Kmin[y][x]);

	A -= alpha[y][x]*B[y][x]*Kmin[y][x];

	if ( lappr1 - A*sq_dist1/2 > (Kmin[y][x]+ THRESH) &&
		 lappr2 - A*sq_dist2/2 > (Kmin[y][x]+ THRESH) &&
		 (Kmin[y][x] < -T) )	        
		    data->setType(y, x, RIDGE);      
	 }

  for (y=0;y < ROW; y++){ 
	  delete [] B[y];
	 }
  delete [] B;
}

void Featurer::non_max_suppr( float **Kmax, float **Kmin, float (**Vmax )[2], 
							 float (**Vmin )[2], float** alpha, float T)
{
  int x,y;
  float tg, lappr1, lappr2;
  float d, sq_dist1, sq_dist2, A; 

  int ROW = data->row;
  int COL = data->col;

  float **B = new float*[ROW];
  for(int i=0; i<ROW; i++){
		B[i] = new float[COL];	
		for (int j=0; j<COL; j++)
			B[i][j] =0.0;
	}
 
  for (y = 1; y < ROW - 1; y++)
    for (x = 1; x < COL - 1; x++)
    {	
	  if (fabs(Vmin[y][x][0]) < eps) {
	      lappr1 = Kmin[y+1][x]; lappr2 = Kmin[y-1][x];
		  d = 1;
	  }
	  else {
	      tg = Vmin[y][x][1] / Vmin[y][x][0];

	      if ((tg >= 0.) && (tg < 1.)) {
	         lappr1 = Kmax[y+1][x+1] * (1.- tg) + tg * Kmax[y][x+1];
	         lappr2 = Kmax[y-1][x-1] * (1.- tg) + tg * Kmax[y][x-1];
			 d = sqrt(1+tg*tg);
		  }
          else if ( tg >= 1.) {
	         lappr1 = (Kmax[y+1][x+1] * (tg - 1.) + Kmax[y+1][x]) / tg;
	         lappr2 = (Kmax[y-1][x-1] * (tg - 1.) + Kmax[y-1][x]) / tg;
             d = sqrt(1+1/tg*tg);
		  }
          else if ((tg <= 0.) && (tg >  -1.)) {
	         lappr1 = Kmax[y+1][x-1] * (1.+ tg) - tg * Kmax[y][x-1];
	         lappr2 = Kmax[y-1][x+1] * (1.+ tg) - tg * Kmax[y][x+1];
             d = sqrt(1+tg*tg);
		  }
	     else if ( tg <= -1.) {
	         lappr1 = (Kmax[y+1][x-1] * (tg + 1.) - Kmax[y+1][x]) / tg;
	         lappr2 = (Kmax[y-1][x+1] * (tg + 1.) - Kmax[y-1][x]) / tg;
             d = sqrt(1+1/tg*tg);
		 }
	  }

	  B[y][x] = (lappr1 - lappr2)/(2*d);
	  if(Vmax[y][x][0]*Vmax[y][x][1] - Vmin[y][x][0]*Vmin[y][x][1]< 0)
      {
        B[y][x] = -(lappr1 - lappr2)/(2*d);
      }  
	}


	for (y = 1; y < ROW - 1; y++)
     for (x = 1; x < COL - 1; x++)
     {	
		if ( Geometry::isZero(Vmax[y][x][0]) ){
	      lappr1 = Kmax[y+1][x]; lappr2 = Kmax[y-1][x];
          sq_dist1 = 1+SQ(data->getValue(y+1,x) - data->getValue(y,x));
          sq_dist2 = 1+SQ(data->getValue(y-1,x) - data->getValue(y,x)); 
	}
	else {
	  tg = Vmax[y][x][1] / Vmax[y][x][0];

	  if ((tg >= 0.) && (tg < 1.)) {
	    lappr1 = Kmax[y+1][x+1] * (1.- tg) + tg * Kmax[y][x+1];
	    lappr2 = Kmax[y-1][x-1] * (1.- tg) + tg * Kmax[y][x-1];
          sq_dist1 = 1+tg*tg + SQ(data->getValue(y+1,x+1) * (1.- tg) 
                     + tg * data->getValue(y,x+1) - data->getValue(y,x));
          sq_dist2 = 1+tg*tg + SQ(data->getValue(y-1,x-1) * (1.- tg) 
                     + tg * data->getValue(y,x-1) - data->getValue(y,x));
	  }
	  else if ( tg >= 1.) {
	    lappr1 = (Kmax[y+1][x+1] * (tg - 1.) + Kmax[y+1][x]) / tg;
	    lappr2 = (Kmax[y-1][x-1] * (tg - 1.) + Kmax[y-1][x]) / tg;
          sq_dist1 = 1+1/(tg*tg)+ SQ(data->getValue(y+1,x+1) * (tg - 1.)/tg
                             + data->getValue(y+1,x)/tg - data->getValue(y,x));
          sq_dist2 = 1+1/(tg*tg)+ SQ(data->getValue(y-1,x-1) * (tg - 1.)/tg
                              + data->getValue(y-1,x)/tg - data->getValue(y,x));
	  }

	  else if ((tg <= 0.) && (tg >  -1.)) {
 	    lappr1 = Kmax[y+1][x-1] * (1.+ tg) - tg *Kmax[y][x-1];
	    lappr2 = Kmax[y-1][x+1] * (1.+ tg) - tg * Kmax[y][x+1];
          sq_dist1 = 1+tg*tg + SQ(data->getValue(y+1,x-1) * (1.+ tg) 
                     - tg * data->getValue(y,x-1) - data->getValue(y,x));
          sq_dist2 = 1+tg*tg + SQ(data->getValue(y-1,x+1) * (1.+ tg) 
                      - tg * data->getValue(y,x+1) - data->getValue(y,x));
	  }
	  else if ( tg <= -1.) {
	    lappr1 = (Kmax[y+1][x-1] * (tg + 1.) - Kmax[y+1][x]) / tg;
	    lappr2 = (Kmax[y-1][x+1] * (tg + 1.) - Kmax[y-1][x]) / tg;
          sq_dist1 = 1+1/(tg*tg)+ SQ(data->getValue(y+1,x-1) * (tg + 1.)/tg
                            - data->getValue(y+1,x)/tg - data->getValue(y,x));
          sq_dist2 = 1+1/(tg*tg)+SQ(data->getValue(y-1,x+1) * (tg + 1.)/tg
                            - data->getValue(y-1,x)/tg - data->getValue(y,x));
	  }
	}
        
	A = 0;
    if ( (Kmax[y][x]-Kmin[y][x]) > 0.0001 )
         A = SQ(B[y][x])/(Kmax[y][x]-Kmin[y][x]);

	A -= alpha[y][x]*B[y][x]*Kmax[y][x];

	if ( Kmax[y][x] >lappr1 + A*sq_dist1/2 + THRESH &&
		 Kmax[y][x] >lappr2 + A*sq_dist2/2 + THRESH &&
		 (Kmax[y][x] > T) )
	        data->setType(y, x, RAVINE);
      
	 }

  for (y=0;y < ROW; y++){ 
	  delete [] B[y];
	 }
  delete [] B;
}

void Featurer::hysteresis(float **K, type t, float thresh)
{
	int col = data->col;
    int row = data->row;

	int i;
	type *in= new type[col*row];
	for ( i=0; i<col*row; i++){
		if ( data->Type[i] == REG )
			in[i] = REG;
		else if (data->Type[i] == t )
		    in[i] = t;
		else 
         in[i] = REG;
	}

	type *out= new type[col*row];
	for ( i=0; i<col*row; i++)
		out[i] = REG;

	for (int y=1; y<row-1; y++)
		for (int x=1; x<col-1; x++){
		  if ( data->getType(y, x) == t && fabs(K[y][x]) > thresh)
			  neighbour(out, y, x, t);
		}


	for (int j=0; j<row; j++)
		for (int i=0; i<col; i++){
			if (data->getType(j,i) == t || data->getType(j,i) == REG)		
                   data->setType(j,i,out[j*col+i]);			
		}
  
	delete [] out;
}

//needed for hysteresis
//
void Featurer::neighbour(type *out, int y, int x, type t)
{
  int i, j;

  out[y*data->col+x] = t;
  data->setType(y, x, REG);

  for (j=-1; j<=1; j++)
     for (i=-1; i<=1; i++){
		  if (data->getType(y+j, x+i) == t){
             out[(y+j)*data->col+(x+i)] = t;
             data->setType(y+j,x+i, REG);
             neighbour(out, y+j, x+i,t);
		  }
	  }
}

void Featurer::setData(Data *in)
{
   data = in;
}

void Featurer::trimm_Kmax(float**Kmax, int T_low, int T_hi, float &T_hyst, float &T_curv)
{
   int n_points = data->col*data->row;

   int valid =0;
   int size;
   if (data->flag != NULL){
	  for (int i=0; i< n_points; i++){
        if (data->flag[i] == 1)
			valid++;
	  }
      size = valid;
	}
   else {
      size = n_points;
   }

  float *tmp = new float[size];
  for (int j=0; j<size; j++)
	  tmp[j] =0;

  int count=0;
  if (data->flag != NULL){
      for (int j=1; j<data->row-1; j++)
	   for (int i=1; i<data->col-1; i++){      
		   if (data->flag[j*data->col+i]){
			   if (Kmax[j][i] >0){
			     tmp[count] = Kmax[j][i];
			     count++;
			   }
		   }
	   }
      
   }
   else{
       for (int j=1; j<data->row-1; j++)
	     for (int i=1; i<data->col-1; i++){
			 if (Kmax[j][i] >0){ 
             tmp[count] = Kmax[j][i];
			 count++;
			 }
	   }
   }
 
    float *tmp1 = new float[count];
	for (int i=0; i<count; i++)
        tmp1[i] = tmp[i];

    qsort(tmp1, count, sizeof(float ), compare);
   
	int pos_low = (int)(T_low*0.01*size);
	int pos_hi = (int)(T_hi*0.01*size);

    T_hyst = tmp[pos_low];
	T_curv = tmp[pos_hi];

	delete [] tmp;
    delete [] tmp1;
}

void Featurer::trimm_Kmin(float**Kmin, int T_low, int T_hi, float &T_hyst, float &T_curv)
{
   int n_points = data->col*data->row;

   int valid =0;
   int size;
   if (data->flag != NULL){
	  for (int i=0; i< n_points; i++){
        if (data->flag[i] == 1)
			valid++;
	  }
      size = valid;
	}
   else {
      size = n_points;
   }

  float *tmp = new float[size];
  for (int j=0; j<size; j++)
	  tmp[j] =0;

  int count =0;
  if (data->flag != NULL){
      for (int j=1; j<data->row-1; j++)
	   for (int i=1; i<data->col-1; i++){      
		   if (data->flag[j*data->col+i]){
			   if (Kmin[j][i] <0){
			      tmp[count] = Kmin[j][i];
			      count++;
			   }
		   }
	   }
      
   }
   else{
       for (int j=1; j<data->row-1; j++)
	     for (int i=1; i<data->col-1; i++){ 
			 if (Kmin[j][i]<0){
               tmp[count] = Kmin[j][i];
			   count++;
			 }
	   }
   }
      
   float *tmp1 = new float[count];
	for (int i=0; i<count; i++)
        tmp1[i] = tmp[i];
    
	qsort(tmp1, count, sizeof(float ), compare);
   
	int pos_low = (int)(T_low*0.01*size);
	int pos_hi = (int)(T_hi*0.01*size);

    T_hyst = tmp[pos_low];
	T_curv = tmp[pos_hi];

	delete [] tmp;
    delete [] tmp1;
}

int compare(const void *v1, const void *v2){

	float x = *(float*) v1;
	float y = *(float*) v2;

	if (x>y)
		return 1;
	else if (x < y)
		return (-1);
	else return(0);
}


float Featurer::getSigma(int y, int x, float (** N)[3])
{	
   float sigma =0.0;
   float tmp[8];

   int count =0;
   for (int j= -1; j<=1; j++)
	 for (int i=-1; i<=1; i++){
        float dot = (float)Geometry::DOT(N[y][x],N[y+j][x+i]);
		if (dot >1)
			dot = 1;
		else if (dot <-1)
			dot = -1;
        float phi = (float)acos(fabs(dot));
		float d;
		if (i==0 || j==0)
			d = 1;
		else
			d =sqrt(2);

        tmp[count] = phi/d;
		count++;
	 }

   qsort(tmp, 8, sizeof(float ), compare);

   int pos = (int)m_expo_percent*0.01*8;

   sigma = tmp[pos];

   return sigma;
}

float Featurer::SQ(float x)
{
  return (x*x);
}

void Featurer::thinning(type Type, float** K)
{
	int i,j, x, y; 
    int tmpX, tmpY;
	float mmax;
	BOOL edited = FALSE;

  	int col = data->col;
    int row = data->row;

	for ( y=2; y<row-2; y++){
	   for ( x=2; x<col-2; x++){
		   if (data->getType(y,x) != Type)
			   continue;
		   switch(get_feature_degree(y, x, Type)){
		   case 0:                     
			 data->setType(y,x, REG);
			 edited = TRUE;
			 break;
			 /*
           case 1:
             for (j=-1; j<=1; j++)    
	           for (i=-1; i<=1; i++){
				   if (data->getType(y+j, x+i) == Type && i!=0 && j!=0){
					   if( get_feature_degree(y+j, x+i, Type) == 1){
                         data->setType(y, x, REG);
						 data->setType(y+j, x+i, REG);
						 edited = TRUE;
						 i=1; 
						 j=1;
						 break;
					   }
				   }
			   }
*/
		   case 1:
			 mmax = fabs (K[y-1][x-1]);
			 tmpX = x-1;
			 tmpY = y-1;
             for (j=-1; j<=1; j++)    
	           for (i=-1; i<=1; i++){
				   if ( fabs(K[y+j][x+i]) > mmax ){
                      mmax = fabs(K[y+j][x+i]);
                      tmpX = x+i;
					  tmpY = y+j;
				   }
			   }
			   if (mmax >0 && data->getType(tmpY, tmpX) != Type){
				   data->setType(tmpY, tmpX, Type);
                   edited = TRUE;
			   }
 			   break;
		   case 2:
           // if out of straight line
			   if ( data->getType(y-1,x-1)==Type && data->getType(y-1,x+1)==Type ){
                   if ( 0.7*fabs(K[y][x]) < fabs(K[y-1][x])){
                     data->setType(y,x,REG);
					 data->setType(y-1, x, Type);
                     edited = TRUE;
				   }
                   break;	
			   }
			   else if (data->getType(y-1, x+1)==Type && data->getType(y+1, x+1)==Type){
                   if ( 0.7*fabs(K[y][x]) < fabs(K[y][x+1]) ){
                     data->setType(y,x,REG);
					 data->setType(y, x+1, Type);
                     edited = TRUE;
				   }
                   break;	
			   }
			   else if ( data->getType(y+1,x-1)==Type && data->getType(y-1, x-1)==Type){
                   if ( 0.7*fabs(K[y][x]) < fabs(K[y][x-1]) ){
                     data->setType(y,x,REG);
					 data->setType(y+1, x, Type);
                     edited = TRUE;
				   }
                   break;	
			   }
/*
           //if adjoining a diagonal edge
			   if ( (data->getType(y,x-1)==Type && data->getType(y-1,x)==Type) ||
				    (data->getType(y-1,x)==Type && data->getType(y,x+1)==Type) ||
                    (data->getType(y,x+1)==Type && data->getType(y+1,x)==Type) ||
                    (data->getType(y+1,x)==Type && data->getType(y,x-1)==Type) )
			   {
                  data->setType(y,x,REG); 
                  edited = TRUE;
				  break;
			   }
		   default:
			   ;
*/
		   }   //switch

		   if (edited){
              if (x<4)
				  x =2;
			  else
				  x -=2;
			  if (y<4)
				  y=2;
			  else
				  y -=2;
			  edited = FALSE;
		   }	
	   }
	}
	
}

int Featurer::get_feature_degree(int y, int x, type Type)
{
   int deg =0;
   for (int j=-1; j<=1; j++)    
	 for (int i=-1; i<=1; i++){
		 if (data->getType(y+j, x+i) == Type)
			 deg++;
	 }

   return (deg-1);
}

void Featurer::blow(int size, type T, int** blowed)
{
	int y, x;
  	int col = data->col;
    int row = data->row;
       
  	for ( y=size; y<row-size; y++){
	   for ( x=size; x<col-size; x++){
		   if ( this->data->getType(y,x) == T ){
                for (int j=-size; j<=size; j++)    
	              for (int i=-size; i<=size; i++){
                    blowed[y+j][x+i] = (int)T;
				  }
		   }            
	   }
	}

}

void Featurer::post_process()
{	
	int i, x, y;

	int** blowed = new int* [data->row];
	for ( i=0; i<data->row; i++){
       blowed[i]= new int[data->col];
       for (int j=0; j<data->col; j++)
		   blowed[i][j] = 0;
	 }

    blow(2, RAVINE, blowed);
	skeletoniaze(RAVINE, blowed);

	 for ( y=0; y<data->row; y++){
		for ( x=0; x<data->col; x++){
			if ( blowed[y][x] == (int) RAVINE)
			  data->setType(y,x, RAVINE);	
			else if (data->getType(y,x) == RAVINE)
              data->setType(y,x, REG);
		}
	 }

	for ( i=0; i<data->row; i++){
       for (int j=0; j<data->col; j++)
		   blowed[i][j] = REG;
	 }

     blow(0, RIDGE, blowed);
//     skeletoniaze(RIDGE, blowed);

	 for (y=0; y<data->row; y++){
		for ( x=0; x<data->col; x++){
			if (data->getType(y, x) == RIDGE){
				if (blowed[y][x] == REG)
					data->setType(y,x,REG);
			}
		}
	 }

	 for (i=0; i<data->row; i++)
		 delete [] blowed[i];
	 delete [] blowed;


}

void Featurer::skeletoniaze(type T, int** in)
{
 	int y, x, i;
  	int col = data->col;
    int row = data->row;

	int** curr = new int* [row];
	for (i=0; i<row; i++){
       curr[i]= new int[col];
	   for (int j=0; j<col; j++)
		   curr[i][j] = in[i][j];
	}

	int** next = new int* [row];
	for (i=0; i<row; i++){
       next[i]= new int[col];
	   for (int j=0; j<col; j++)
		   next[i][j] = 0;
	}

    int count=1;
    while (count >0) {
		count =0;
	  for ( y=1; y<row-1; y++){
	    for ( x=1; x<col-1; x++){
		   int mmin = 1 ;
           for (int j=-1; j<=1; j++){    
	         for (int i=-1; i<=1; i++){
				 if (curr[y+j][x+i] == 0){
                   mmin = 0;
				   j=1;
				   i=1;
				   break;
				 }
				 i++;
				 j++;
			 }
		   }      
           next[y][x] = in[y][x] + mmin;
		}
	  }
      
	  count =0;
	  for (int k=0; k<row; k++)
		  for (int s=0; s<col; s++){
			  if (curr[k][s] != next [k][s])
				  count++;
               curr[k][s] = next [k][s];
			  }

	};  //while

	 for ( y=1; y<row-1; y++){
	   for ( x=1; x<col-1; x++){
		     in[y][x] = (int) REG;
			 BOOL flag = FALSE;
		     for (int j=-1; j<=1; j++)    
	           for (int i=-1; i<=1; i++){
				 if (curr[y][x] < curr[y+j][x+i]){
                   flag = TRUE;
				   j=1;
				   i=1;
				   break;
				 }
				 i++;
				 j++;
			 }
			   if (flag){
				   in[y][x] = (int)T;
				   flag = FALSE;
			   }
	   }
	 }


	for (i=0; i<row; i++){
       delete curr[i];
	   delete next[i];
	}

	delete [] curr;
	delete [] next;

}
