// Image.cpp: implementation of the CImage class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "RangeImage.h"
#include "Image.h"
#include "Normal.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CImage::CImage():row_N(0), col_N(0), max_grey(0)
{
  pixelValue = NULL;
  isFeature = NULL;
}


CImage::CImage(int r, int c, int l)
{
 row_N = r;
 col_N = c;
 max_grey = l;

 pixelValue = new float*[row_N];
 for(int i=0; i<row_N; i++){ 
     pixelValue[i] = new float[col_N];
     for(int j=0; j<col_N; j++)
       pixelValue[i][j] = 0;
 }
}

CImage::~CImage()
{
  erase();
}

void CImage::copy(const CImage &old_image)
{
 row_N = old_image.row_N;
 col_N = old_image.col_N;
 max_grey = old_image.max_grey;

 pixelValue = new float*[row_N];
 for(int i=0; i<row_N; i++) {
     pixelValue[i] = new float[col_N];
     for(int j=0; i<col_N; j++) 
       pixelValue[i][j] = old_image.pixelValue[i][j];
 }
}

void CImage::getImageInfo(int &rows, int &cols, int &lev)
{
   rows = row_N;
   cols = col_N;
   lev = max_grey;
}

void CImage::setImageInfo(int rows, int cols, int lev)
{
  row_N = rows;
  col_N = cols;
  max_grey = lev;
}

void CImage::setPixelVal(int row, int col, float val)
{
  pixelValue[row][col] = val;
}

float CImage::getPixelVal(int row, int col)
{
  return(pixelValue[row][col]);
}

void CImage::erase()
{
	if (pixelValue !=NULL){
      for (int i=0; i<row_N; i++)
		delete [] pixelValue[i];

      delete [] pixelValue;
	}

	if (isFeature != NULL){
      for (int i=0; i<row_N; i++)
		delete [] isFeature[i];
      delete[] isFeature;
	}

 row_N = 0;    
 col_N = 0;
 max_grey = 0;
 
}

int CImage::getSize()
{
  return (row_N*col_N);
}

void CImage::allocate(int r, int c)
{
 row_N = r;
 col_N = c;
  
 pixelValue = new float*[row_N];
 for(int i=0; i<row_N; i++) {
     pixelValue[i] = new float[col_N];
 }
}

void CImage::setFeature(int row, int col, type t)
{
  isFeature[row][col]= t;
}

type CImage::getType(int row, int col)
{
  return( isFeature[row][col]);
}

void CImage::allocateFeatures()
{
	isFeature = new type*[row_N];
	for (int i=0; i<row_N; i++)
       isFeature[i] = new type[col_N];
	
	for( i=0; i<row_N; i++){
		for (int j=0; j<col_N; j++)
          isFeature[i][j] = REG;
	}

}

void CImage::image_to_data(Data *data)
{
	int i, j;
	data->step = 1;
	
	for( i=0; i<row_N; i++){
		for( j=0; j<col_N; j++){
		   data->setValue(i, j, 0.1*getPixelVal(i,j));
		}
	}	
}
