// MeshData.cpp: implementation of the MeshData class
//
//////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "RangeImage.h"
#include "MeshData.h"
#include "Geometry.h"
#include <vector>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#define PI 3.14159265

//////////////////////////////////////////////////////////////////////
// 
//////////////////////////////////////////////////////////////////////

MeshData::MeshData()
{
	face_N = 0;
	vertex_N = 0;

	vertex = NULL;
	face = NULL;
	
	normal = NULL;

}

MeshData::~MeshData()
{
	this->deleteLinkData();

}

void MeshData::setVertexCount(int vertex_N)
{
	if(vertex != NULL)
		delete[] vertex;
	this->vertex_N = vertex_N;
	vertex = new float[vertex_N][3];
}

void MeshData::setFaceCount(int face_N)
{
	if(face != NULL)
		delete[] face;
	this->face_N = face_N;
	face = (int (*)[3])new int[face_N][3];
}

int MeshData::getVertexCount()
{
	return vertex_N;
}

int MeshData::getFaceCount()
{
	return face_N;
}

void MeshData::setVertex(int index, float coordinates[3])
{
	vertex[index][0] = coordinates[0];
	vertex[index][1] = coordinates[1];
	vertex[index][2] = coordinates[2];
}

void MeshData::setFace(int index, int vertex_index[])
{
	face[index][0] = vertex_index[0];
	face[index][1] = vertex_index[1];
	face[index][2] = vertex_index[2];
}

void MeshData::getVertex(int index, float coordinates[3])
{
	coordinates[0] = vertex[index][0];
	coordinates[1] = vertex[index][1];
	coordinates[2] = vertex[index][2];
}

void MeshData::getFace(int index, int vertex_index[])
{
	vertex_index[0] = face[index][0];
	vertex_index[1] = face[index][1];
	vertex_index[2] = face[index][2];
}

void MeshData::normalizeScale()
{
	float max[3], min[3];
	max[0] = max[1] = max[2] = -1000000;
	min[0] = min[1] = min[2] = 1000000;
	for(int i=0; i<vertex_N; i++){
		for(int j=0; j<3; j++){
			if(max[j] < vertex[i][j])
				max[j] = vertex[i][j];
			else if(min[j] > vertex[i][j])
				min[j] = vertex[i][j];
		}
	}
	float sizeX = max[0] - min[0];
	float sizeY = max[1] - min[1];
	float sizeZ = max[2] - min[2];

	float scale = sizeX;
	if(scale < sizeY)
		scale = sizeY;
	if(scale < sizeZ)
		scale = sizeZ;
    normalize_scale = scale*0.05;
	scale = 20.0f/scale; //20.0f scale;

	for(i=0; i<vertex_N; i++){
		vertex[i][0] = (vertex[i][0]-0.5f*(max[0]+min[0]))*scale;
		vertex[i][1] = (vertex[i][1]-0.5f*(max[1]+min[1]))*scale;
		vertex[i][2] = (vertex[i][2]-0.5f*(max[2]+min[2]))*scale;
	}

	original_max[0] = max[0];
	original_max[1] = max[1];
	original_max[2] = max[2];

	original_min[0] = min[0];
	original_min[1] = min[1];
	original_min[2] = min[2];


}

void MeshData::computeNormal()
{
	if(normal != NULL)
		delete[] normal;
	normal = new float[vertex_N][3];

	double (*tmp)[3] = new double[vertex_N][3];
	for(int i=0; i<vertex_N; i++)
		tmp[i][0] = tmp[i][1] = tmp[i][2] = 0;

	for(i=0; i<face_N; i++){
		int *f = face[i];
		double area = Geometry::AREA(vertex[f[0]], vertex[f[1]], vertex[f[2]]);
		float v1[3], v2[3];
		double c[3];
		Geometry::VEC(v1, vertex[f[1]], vertex[f[0]]);
		Geometry::VEC(v2, vertex[f[2]], vertex[f[0]]);
		Geometry::CROSS(c, v1, v2);

		tmp[f[0]][0] += c[0];
		tmp[f[0]][1] += c[1];
		tmp[f[0]][2] += c[2];

		tmp[f[1]][0] += c[0];
		tmp[f[1]][1] += c[1];
		tmp[f[1]][2] += c[2];

		tmp[f[2]][0] += c[0];
		tmp[f[2]][1] += c[1];
		tmp[f[2]][2] += c[2];
	}

	for(i=0; i<vertex_N; i++){
		double length = Geometry::LENGTH(tmp[i]);
		if((float)length != 0){
			normal[i][0] = (float)(tmp[i][0]/length);
			normal[i][1] = (float)(tmp[i][1]/length);
			normal[i][2] = (float)(tmp[i][2]/length);
		}
		else
			normal[i][0] = normal[i][1] = normal[i][2] = 0.0f;
	}

	delete[] tmp;
}

double MeshData::getVolume()
{
	double vol = 0;
	double g[3],n[3],v1[3],v2[3];

	for(int i=0; i<face_N; i++){
		g[0] = (vertex[ face[i][0] ][0] + vertex[ face[i][1] ][0] + vertex[ face[i][2] ][0])/3.0;
		g[1] = (vertex[ face[i][0] ][1] + vertex[ face[i][1] ][1] + vertex[ face[i][2] ][1])/3.0;
		g[2] = (vertex[ face[i][0] ][2] + vertex[ face[i][1] ][2] + vertex[ face[i][2] ][2])/3.0;

		v1[0] = vertex[ face[i][1] ][0] - vertex[ face[i][0] ][0];
		v1[1] = vertex[ face[i][1] ][1] - vertex[ face[i][0] ][1];
		v1[2] = vertex[ face[i][1] ][2] - vertex[ face[i][0] ][2];

		v2[0] = vertex[ face[i][2] ][0] - vertex[ face[i][0] ][0];
		v2[1] = vertex[ face[i][2] ][1] - vertex[ face[i][0] ][1];
		v2[2] = vertex[ face[i][2] ][2] - vertex[ face[i][0] ][2];

		n[0] = v1[1]*v2[2] - v1[2]*v2[1];
		n[1] = v1[2]*v2[0] - v1[0]*v2[2];
		n[2] = v1[0]*v2[1] - v1[1]*v2[0];

		vol += (g[0]*n[0] + g[1]*n[1] + g[2]*n[2])/6.0;
	}
	return vol;
}


void MeshData::copyMesh(MeshData *mesh)
{
	setVertexCount(mesh->vertex_N);
	for(int i=0; i<vertex_N; i++){
		vertex[i][0] = mesh->vertex[i][0];
		vertex[i][1] = mesh->vertex[i][1];
		vertex[i][2] = mesh->vertex[i][2];
	}

	setFaceCount(mesh->face_N);
	for(i=0; i<face_N; i++){
		face[i][0] = mesh->face[i][0];
		face[i][1] = mesh->face[i][1];
		face[i][2] = mesh->face[i][2];
	}
}

void MeshData::rescale(double rate)
{
	for(int i=0; i<vertex_N; i++){
		vertex[i][0] = (float)(rate*vertex[i][0]);
		vertex[i][1] = (float)(rate*vertex[i][1]);;
		vertex[i][2] = (float)(rate*vertex[i][2]);
	}
}

void MeshData::deleteLinkData()
{
	if(vertex != NULL)
		delete[] vertex;

	if(face != NULL)
		delete[] face;

	if(normal != NULL)
		delete[] normal;

	face_N = 0;
	vertex_N = 0;

	vertex = NULL;
	face = NULL;

    normal = NULL;

}

void MeshData::faceCenter(double c[], int f)
{
	int i1 = face[f][0];
	int i2 = face[f][1];
	int i3 = face[f][2];
	c[0] = (vertex[i1][0] + vertex[i2][0] + vertex[i3][0])/3.0;
	c[1] = (vertex[i1][1] + vertex[i2][1] + vertex[i3][1])/3.0;
	c[2] = (vertex[i1][2] + vertex[i2][2] + vertex[i3][2])/3.0;
}

double MeshData::faceArea(int f)
{
	return Geometry::AREA(vertex[face[f][0]], vertex[face[f][1]], vertex[face[f][2]]);
}

void MeshData::shift(float v[3])
{
	for(int i=0; i<vertex_N; i++){
		vertex[i][0] = vertex[i][0]+v[0];
		vertex[i][1] = vertex[i][1]+v[1];
		vertex[i][2] = vertex[i][2]+v[2];
	}
}

void MeshData::getCentroid(float p[])
{
	
	p[0]=p[1]=p[2]=0;

	for(int i=0; i<vertex_N; i++){
		p[0] += this->vertex[i][0];
		p[1] += this->vertex[i][1];
		p[2] += this->vertex[i][2];
	}

	p[0] /= vertex_N;
	p[1] /= vertex_N;
	p[2] /= vertex_N;

}

void MeshData::undoNormalizeScale()
{
	for(int i=0; i<vertex_N; i++){
		vertex[i][0] = vertex[i][0]*normalize_scale + 0.5f*(original_max[0] + original_min[0]);
		vertex[i][1] = vertex[i][1]*normalize_scale + 0.5f*(original_max[1] + original_min[1]);
		vertex[i][2] = vertex[i][2]*normalize_scale + 0.5f*(original_max[2] + original_min[2]);
	}
}

//return number of triangles containing that edge
//


void MeshData::reverseTr(int trId)
{
  int temp0 = this->face[trId][0];

  this->face[trId][0] = this->face[trId][2];
  this->face[trId][2] = temp0;
}




