// MeshData.h: MeshData 
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_MESHDATA_H__958F025D_682B_4971_A477_DC00A2379F28__INCLUDED_)
#define AFX_MESHDATA_H__958F025D_682B_4971_A477_DC00A2379F28__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "math.h"

#define INNER 0
#define BOUNDARY 1
#define NON_MANIFOLD 2

class MeshData  
{
public:

	//The number of vertex and face (File data)
	int face_N, vertex_N;

	//Vertex coordinates
	float (*vertex)[3];
	//Face index
	int (*face)[3];

	//Vertex normal
	float (*normal)[3];

	//Bounding Box Data for Normalize Model Size
	float original_max[3], original_min[3];
	double normalize_scale;

private:

public:
	void reverseTr(int trId);
	void getCentroid(float p[3]);
	void shift(float v[3]);
	double faceArea(int f);
	void faceCenter(double c[3], int f);
	void deleteLinkData();
	void rescale(double rate);
	void copyMesh(MeshData* mesh);
	double getVolume();
	void computeNormal();

	//Scale mesh size into 20x20x20 cube 
	void normalizeScale();
    //unscale mesh
	void undoNormalizeScale();

	//get & set data (These data can be acessed directly, so no need)
	void getFace(int index, int vertex_index[3]);
	void getVertex(int index, float coordinates[3]);
	void setFace(int index, int vertex_index[3]);
	void setVertex(int index, float coordinates[3]);
	int getFaceCount();
	int getVertexCount();
	void setFaceCount(int face_N);
	void setVertexCount(int vertex_N);

	MeshData();
	virtual ~MeshData();

};

#endif // !defined(AFX_MESHDATA_H__958F025D_682B_4971_A477_DC00A2379F28__INCLUDED_)
