// Normal.cpp: implementation of the Normal class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "RangeImage.h"
#include "Normal.h"
#include "Geometry.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#include <math.h>
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Normal::Normal()
{
  normal_v = NULL;
  int rows = 0;
  int cols = 0;
}

Normal::~Normal()
{
	if(normal_v != NULL)
        delete [] normal_v;
	
	cols = 0;
	rows = 0;
}

void Normal::get_normal_v(Data *dt)
{
	int i, j;

	rows = dt->row;
	cols = dt->col;

	normal_v = new float[rows*cols][3];

	for( j=1; j<rows-1; j++)
    	for( i=1; i<cols-1; i++){
			float Ix = 0.5f*(dt->getValue(j,i+1) - dt->getValue(j,i-1));
			float Iy = 0.5f*(dt->getValue(j+1,i) - dt->getValue(j-1,i));

			double len = sqrt(Ix*Ix + Iy*Iy + 1.0);

			normal_v[j*cols+i][0] = (float)(Ix/len);
			normal_v[j*cols+i][1] = (float)(Iy/len);
			normal_v[j*cols+i][2] = (float)(-1.0/len);
		}
}

void Normal::normal_to_mesh(Data *dt)
{
    if (dt->Faces == NULL)
		return;

	int rows = dt->row;
	int cols = dt->col;

	if(normal_v  != NULL)
		delete[] normal_v;
	normal_v = new float[rows*cols][3];

	double (*tmp)[3] = new double[rows*cols][3];
	for(int i=0; i<rows*cols; i++)
		tmp[i][0] = tmp[i][1] = tmp[i][2] = 0;

	for(i=0; i<2*(rows-1)*(cols-1); i++){
		int *f = dt->Faces[i];
//		double area = Geometry::AREA(dt->Points[f[0]],dt->Points[f[1]], dt->Points[f[2]]);
		float v1[3], v2[3];
		double c[3];

		Geometry::VEC(v1, dt->Points[f[1]], dt->Points[f[0]]);
		Geometry::VEC(v2, dt->Points[f[2]], dt->Points[f[0]]);
		Geometry::CROSS(c, v1, v2);

		tmp[f[0]][0] += c[0];
		tmp[f[0]][1] += c[1];
		tmp[f[0]][2] += c[2];

		tmp[f[1]][0] += c[0];
		tmp[f[1]][1] += c[1];
		tmp[f[1]][2] += c[2];

		tmp[f[2]][0] += c[0];
		tmp[f[2]][1] += c[1];
		tmp[f[2]][2] += c[2];
	}

	for(i=0; i<rows*cols; i++){
		double length = Geometry::LENGTH(tmp[i]);
		if((float)length != 0){
			normal_v[i][0] = (float)(tmp[i][0]/length);
			normal_v[i][1] = (float)(tmp[i][1]/length);
			normal_v[i][2] = (float)(tmp[i][2]/length);
		}
		else
			normal_v[i][0] = normal_v[i][1] = normal_v[i][2] = 0.0f;
	}

	delete [] tmp;
}
