// RangeImageView.cpp : implementation of the CRangeImageView class
//

#include "stdafx.h"
#include "RangeImage.h"
#include "Image.h"
#include "Normal.h"
#include "Cloud.h"
#include "Geometry.h"

#include "RangeImageDoc.h"
#include "RangeImageView.h"


#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CRangeImageView

IMPLEMENT_DYNCREATE(CRangeImageView, CView)

BEGIN_MESSAGE_MAP(CRangeImageView, CView)
	//{{AFX_MSG_MAP(CRangeImageView)
	ON_WM_CREATE()
	ON_WM_DESTROY()
	ON_WM_ERASEBKGND()
	ON_WM_SIZE()
	ON_WM_MOUSEMOVE()
	ON_WM_KEYDOWN()
	ON_WM_LBUTTONUP()
	ON_WM_LBUTTONDOWN()
	ON_WM_RBUTTONDOWN()
	ON_WM_RBUTTONUP()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CRangeImageView construction/destruction

void sphere(float r, float p[], int n);

CRangeImageView::CRangeImageView()
{
	// TODO: add construction code here
	isPerspective = true;

	eye[0] = 126;
	eye[1] = -148;
	eye[2] = 140;

	center[0] = 0.0;
	center[1] = 0.0;
	center[2] = 0.0;

	Rx = 0;
	Ry = 0;
	Rz = 0;

	vx[0] = 0;
	vx[1] = 0;
	vx[2] = 1;

	vy[0] = -(float)(eye[1]/sqrt(eye[0]*eye[0]+eye[1]*eye[1]));
	vy[1] = (float)(eye[0]/sqrt(eye[0]*eye[0]+eye[1]*eye[1]));
	vy[2] = 0;

	view_angle = 10.5;

    isLPressed = FALSE;
	isRPressed = FALSE;

	shift_x = 0;
	shift_y = 0;
	shift_z = 0;

	m_back_plane_on = TRUE;
}

CRangeImageView::~CRangeImageView()
{

}

BOOL CRangeImageView::PreCreateWindow(CREATESTRUCT& cs)
{
	// TODO: Modify the Window class or styles here by modifying
	//  the CREATESTRUCT cs

	return CView::PreCreateWindow(cs);
}

/////////////////////////////////////////////////////////////////////////////
// CRangeImageView drawing

void CRangeImageView::OnDraw(CDC* pDC)
{
	CRangeImageDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);
	// TODO: add draw code for native data here

	wglMakeCurrent(pDC->m_hDC, m_hRC);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glPushMatrix();
	glTranslatef(shift_x,shift_y,shift_z);
	glRotatef(Ry,vy[0],vy[1],vy[2]);
	glRotatef(Rz,vx[0],vx[1],vx[2]);
	pDoc->drawImage();
	glPopMatrix();

	glFlush();
	SwapBuffers(wglGetCurrentDC());
	wglMakeCurrent(pDC->m_hDC, NULL);
}

/////////////////////////////////////////////////////////////////////////////
// CRangeImageView diagnostics

#ifdef _DEBUG
void CRangeImageView::AssertValid() const
{
	CView::AssertValid();
}

void CRangeImageView::Dump(CDumpContext& dc) const
{
	CView::Dump(dc);
}

CRangeImageDoc* CRangeImageView::GetDocument() // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CRangeImageDoc)));
	return (CRangeImageDoc*)m_pDocument;
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CRangeImageView message handlers

int CRangeImageView::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	if (CView::OnCreate(lpCreateStruct) == -1)
		return -1;
	
	// TODO: Add your specialized creation code here
		PIXELFORMATDESCRIPTOR pfd =
	{
		sizeof(PIXELFORMATDESCRIPTOR),
		1,
		PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
		PFD_TYPE_RGBA,
		16,
		0,0,0,0,0,0,
		0,0,0,0,0,0,0,
		32,
		0,0,
		PFD_MAIN_PLANE,
		0,
		0,0,0
	};

	CClientDC  clientDC(this);

	int pixelFormat = ChoosePixelFormat(clientDC.m_hDC, &pfd);
	if(!SetPixelFormat(clientDC.m_hDC, pixelFormat, &pfd))
		return -1;

	m_hRC = wglCreateContext(clientDC.m_hDC);

	wglMakeCurrent(clientDC.m_hDC,m_hRC);
	initGL();
	wglMakeCurrent(clientDC.m_hDC,NULL);
	return 0;
}

void CRangeImageView::initGL()
{
 
	GLfloat ambient[4] = {0.8f,0.8f,0.8f,1.0f};

	GLfloat material_color[4] = {0.7, 0.7, 0.7, 0.0};
	GLfloat material_specular[4] = {0.6,0.6,0.6,0.8};
	GLfloat material_ambient[4] = {0.7,0.7,0.7,0.0};

	glClearColor(1,1,1,0);

	glLightModelfv(GL_LIGHT_MODEL_TWO_SIDE , ambient);
  
//glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, material_specular);
//	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, material_color);
//	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, material_ambient);
//	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 128);

	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LESS);

}



void CRangeImageView::OnDestroy() 
{
	CView::OnDestroy();
	
	// TODO: Add your message handler code here
  if(wglGetCurrentContext()!=NULL) 
   wglMakeCurrent(NULL, NULL) ;
   
    if(m_hRC!=NULL)
        {
      wglDeleteContext(m_hRC);
      m_hRC = NULL;
    }

 }

void CRangeImageView::changeView()
{
	CClientDC* dc = new CClientDC(this); 
	OnDraw(dc);
	delete dc;
}


BOOL CRangeImageView::OnEraseBkgnd(CDC* pDC) 
{
	// TODO: Add your message handler code here and/or call default
	
	return TRUE;
} 

void CRangeImageView::OnSize(UINT nType, int cx, int cy) 
{
	CView::OnSize(nType, cx, cy);
	
	// TODO: Add your message handler code here
	int i;
	GLdouble up[3];
	GLdouble vector[3];
	GLdouble norm;

	CClientDC clientDC(this);

	wglMakeCurrent(clientDC.m_hDC, m_hRC);

	glViewport(0, 0, cx, cy);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	//gluPerspective(3, (GLfloat)cx/(GLfloat)cy, 10.0, 4000.0);
	if(isPerspective)
		gluPerspective(view_angle, (GLfloat)cx/(GLfloat)cy, 10.0, 4000.0);
	else
		glOrtho(-10,10,-10.0*cy/cx,10.0*cy/cx,10,4000);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	for(i=0; i<3; i++)
		vector[i] = center[i] - eye[i];

	up[0] =  vector[0] * vector[2];
	up[1] =  vector[1] * vector[2];
	up[2] =  vector[0] * vector[0] + vector[1] * vector[1];

	norm = up[0] * up[0] + up[1] * up[1] + up[2] * up[2];
	norm = sqrt(norm);
	for(i=0; i<3; i++)
		up[i] = up[i] / norm;
	
	gluLookAt(eye[0], eye[1], eye[2], center[0], center[1], center[2],
		up[0], up[1], up[2]);

	wglMakeCurrent(clientDC.m_hDC, NULL);	
}

void CRangeImageView::OnMouseMove(UINT nFlags, CPoint point) 
{
	// TODO: Add your message handler code here and/or call default
	if(isLPressed){
		int nx = point.x;
		int ny = point.y;
		
		Rz += (nx - px)/3;
		Ry += (ny - py)/3;
		px = nx;
		py = ny;
		CClientDC* dc = new CClientDC(this); 
		OnDraw(dc);
		delete dc;
	}
	else if(isRPressed){
		int nx = point.x;
		int ny = point.y;

		view_angle += (ny - py)/20.0f;
		px = nx;
		py = ny;
		CClientDC* dc = new CClientDC(this);

		RECT* rect = new RECT;
		GetClientRect(rect);
		OnSize(SIZE_RESTORED, rect->right, rect->bottom);
		delete rect;
		OnDraw(dc);
		delete dc;
	}	
	CView::OnMouseMove(nFlags, point);
}

void CRangeImageView::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	// TODO: Add your message handler code here and/or call default
	switch(nChar){
	case VK_UP:
		shift_z += 1;
		break;
	case VK_DOWN:
		shift_z -= 1;
		break;
	case VK_RIGHT:
		shift_x += 1;
		break;
	case VK_LEFT:
		shift_x -= 1;
		break;
	}
	CClientDC* dc = new CClientDC(this); 
	OnDraw(dc);
	delete dc;
	
	CView::OnKeyDown(nChar, nRepCnt, nFlags);
}

void CRangeImageView::OnLButtonUp(UINT nFlags, CPoint point) 
{
	// TODO: Add your message handler code here and/or call default
	isLPressed = false;

	CView::OnLButtonUp(nFlags, point);
}

void CRangeImageView::OnLButtonDown(UINT nFlags, CPoint point) 
{
	// TODO: Add your message handler code here and/or call default
    isLPressed = true;
	px = point.x;
	py = point.y;

	CView::OnLButtonDown(nFlags, point);	
}

void CRangeImageView::OnRButtonUp(UINT nFlags, CPoint point) 
{
	// TODO: Add your message handler code here and/or call default
	isRPressed = false;

	CView::OnRButtonUp(nFlags, point);
}

void CRangeImageView::OnRButtonDown(UINT nFlags, CPoint point) 
{
	// TODO: Add your message handler code here and/or call default
	isRPressed = true;
	px = point.x;
	py = point.y;

	CView::OnRButtonDown(nFlags, point);	
}

void CRangeImageView::drawShaded(Data *dt)
{
	int cols = dt->col;
	int	rows = dt->row;
    float step = dt->step;

	float c_x = 0.5*step*cols;
    float c_y = 0.5*step*rows;

	Normal nl;
	nl.get_normal_v(dt);

    glShadeModel (GL_FLAT);   

	float n[3];
    for (int j=0; j<rows-1; j++){	  
	   for (int i=0; i<cols-1; i++){
 			glBegin(GL_TRIANGLES);

            int curr = j*cols+i;			
			n[0] = nl.normal_v[curr][0];
			n[1] = nl.normal_v[curr][1];
			n[2] = nl.normal_v[curr][2];
			float len = sqrt(n[0]*n[0] + n[1]*n[1] + n[2]*n[2]);
			n[0] /= len;
			n[1] /= len;
			n[2] /= len;
			glNormal3fv(n);
			glVertex3f(i*step-c_x, j*step-c_y, dt->Value[curr]);

            curr = (j+1)*cols+i;
			n[0] = nl.normal_v[curr][0];
			n[1] = nl.normal_v[curr][1];
			n[2] = nl.normal_v[curr][2];
			len = sqrt(n[0]*n[0] + n[1]*n[1] + n[2]*n[2]);
			n[0] /= len;
			n[1] /= len;
			n[2] /= len;
			glNormal3fv(n);
			glVertex3f(i*step-c_x, (j+1)*step-c_y, dt->Value[curr]);

            curr = (j+1)*cols+(i+1);
			n[0] = nl.normal_v[curr][0];
			n[1] = nl.normal_v[curr][1];
			n[2] = nl.normal_v[curr][2];
			len = sqrt(n[0]*n[0] + n[1]*n[1] + n[2]*n[2]);
			n[0] /= len;
			n[1] /= len;
			n[2] /= len;
			glNormal3fv(n);			
			glVertex3f((i+1)*step-c_x, (j+1)*step-c_y, dt->Value[curr]);


			//the second triangle
            // 
			curr = j*cols+i;			
			n[0] = nl.normal_v[curr][0];
			n[1] = nl.normal_v[curr][1];
			n[2] = nl.normal_v[curr][2];

			len = sqrt(n[0]*n[0] + n[1]*n[1] + n[2]*n[2]);
			n[0] /= len;
			n[1] /= len;
			n[2] /= len;
			glNormal3fv(n);
			glVertex3f(i*step-c_x, j*step-c_y, dt->Value[curr]);

			curr = (j+1)*cols+(i+1);
			n[0] = nl.normal_v[curr][0];
			n[1] = nl.normal_v[curr][1];
			n[2] = nl.normal_v[curr][2];
			len = sqrt(n[0]*n[0] + n[1]*n[1] + n[2]*n[2]);
			n[0] /= len;
			n[1] /= len;
			n[2] /= len;
			glNormal3fv(n);
			glVertex3f((i+1)*step-c_x, (j+1)*step-c_y, dt->Value[curr]);


			curr = j*cols+(i+1);
			n[0] = nl.normal_v[curr][0];
			n[1] = nl.normal_v[curr][1];
			n[2] = nl.normal_v[curr][2];
			len = sqrt(n[0]*n[0] + n[1]*n[1] + n[2]*n[2]);
			n[0] /= len;
			n[1] /= len;
			n[2] /= len;
			glNormal3fv(n);
			glVertex3f((i+1)*step-c_x, j*step-c_y, dt->Value[curr]);

			glEnd();
			
		}
	}
}

void CRangeImageView::drawFeatures(Data *dt)
{
	int cols = dt->col;
	int	rows = dt->row;
    float step = dt->step;

	float c_x = 0.5*step*cols;
    float c_y = 0.5*step*rows;

	GLUquadricObj *plainQuad;
	plainQuad = gluNewQuadric(); 
    gluQuadricDrawStyle(plainQuad, GLU_FILL); 
    gluQuadricOrientation(plainQuad, GLU_OUTSIDE);

	glShadeModel(GL_FLAT);
	glDisable(GL_LIGHTING);

	glPointSize(5);
	glBegin(GL_POINTS);

//to avoid "false"features move avay plane

	for(int y=0; y<rows; y++)
	  for(int x=0; x<cols; x++){
          if( dt->getType(y,x) == REG ||Geometry::isZero(0.1*dt->getValue(y,x)) )
			  continue;
		     float c[3] = {step*x-c_x, step*y-c_y, dt->getValue(y,x)};
             if(dt->getType(y,x) == RIDGE){
 				glColor3f(0,0,1);
				sphere(0.7*step, c, 6);
		   }
		  else if(dt->getType(y,x) == RAVINE){
                glColor3f(1,0,0);
				sphere(0.7*step, c, 6);
		  }
	   }

    glEnd();
	glEnable(GL_LIGHTING);
    gluDeleteQuadric(plainQuad);
}


void CRangeImageView::drawHeigh()
{
	glShadeModel(GL_FLAT);
	glDisable(GL_LIGHTING);

	glPointSize(4);
	glColor3f(1,0,0);
/*
    float x_min, x_max, y_min, y_max;
	x_min = cl->data[0].x;
    x_max = cl->data[0].x;
    y_max = cl->data[0].y;
    y_min = cl->data[0].y;

	for (int i=1; i<cl->size; i++){
      if (cl->data[i].x < x_min)
         x_min = cl->data[i].x;
	  if (cl->data[i].x > x_max)
         x_max = cl->data[i].x;

	  if (cl->data[i].y < y_min)
         y_min = cl->data[i].y;
	  if (cl->data[i].y > y_max)
         y_max = cl->data[i].y;
	}

	float c_x = 0.5f*(x_max-x_min);
	float c_y = 0.5f*(y_max-y_min);

	glBegin(GL_POINTS);			  
	
	for( i=0; i<cl->size; i++)
	   glVertex3f(cl->data[i].x - c_x, cl->data[i].y - c_y, cl->data[i].height);
		   
    glEnd();
	glEnable(GL_LIGHTING);
	*/
}

void CRangeImageView::drawPoints(Data *data)
{

	float c_x = 0.5*data->step*(data->col-1);
    float c_y = 0.5*data->step*(data->row-1);

	glShadeModel(GL_FLAT);
	glDisable(GL_LIGHTING);

	glPointSize(4);
	glColor3f(1,0,0);
	glBegin(GL_POINTS);			  	
	  for (int j=0; j<data->row-1; j++){	  
	     for (int i=0; i<data->col-1; i++){
          glVertex3f(i*data->step-c_x, j*data->step-c_y, data->Value[j*data->col+i]);
		 }
	  }

    glEnd();
	glEnable(GL_LIGHTING);
}

void CRangeImageView::drawMesh(Data *dt)
{
  if (dt->Faces == NULL)
	  dt->triangulate();

  float c_x = 0.5*dt->step*dt->col;
  float c_y = 0.5*dt->step*dt->row;


    Normal nl;
	nl.normal_to_mesh(dt);

    glShadeModel (GL_FLAT);   
    int nTrs = 2*(dt->row-1)*(dt->col-1);

	if (dt->flag == NULL || m_back_plane_on){ 
       for (int j=0; j<nTrs; j++){	  

 			glBegin(GL_TRIANGLES);
            
			glNormal3fv(nl.normal_v[dt->Faces[j][0]]);
			glVertex3fv(dt->Points[dt->Faces[j][0]]);

			glNormal3fv(nl.normal_v[dt->Faces[j][1]]);
			glVertex3fv(dt->Points[dt->Faces[j][1]]);

			glNormal3fv(nl.normal_v[dt->Faces[j][2]]);
			glVertex3fv(dt->Points[dt->Faces[j][2]]);

			glEnd();
	   }
	}  else {
        for (int j=0; j<nTrs; j++){	  
            if (dt->flag[dt->Faces[j][0]] + dt->flag[dt->Faces[j][1]]+dt->flag[dt->Faces[j][2]] <3 )
				continue;

			glBegin(GL_TRIANGLES);

			glNormal3fv(nl.normal_v[dt->Faces[j][0]]);
			glVertex3fv(dt->Points[dt->Faces[j][0]]);

			glNormal3fv(nl.normal_v[dt->Faces[j][1]]);
			glVertex3fv(dt->Points[dt->Faces[j][1]]);

			glNormal3fv(nl.normal_v[dt->Faces[j][2]]);
			glVertex3fv(dt->Points[dt->Faces[j][2]]);

			glEnd();
		}
	}
}

void sphere(float r, float p[], int n)
{
	float (**vertex)[3] = (float(**)[3])new float[n];
	float (**normal)[3] = (float(**)[3])new float[n];
	for(int i=0; i<n; i++){
		vertex[i] = (float (*)[3])new float[n][3];
		normal[i] = (float (*)[3])new float[n][3];
	}

	for(i=0; i<n; i++){
		float u = (float)(PI*i/(n-1) - 0.5f*PI);
		for(int j=0; j<n; j++){
			float v = (float)(2*PI*j/(n-1));
			normal[i][j][0] = (float)(cos(v)*cos(u));
			normal[i][j][1] = (float)(sin(v)*cos(u));
			normal[i][j][2] = (float)sin(u);

			vertex[i][j][0] = (float)(r*normal[i][j][0] + p[0]);
			vertex[i][j][1] = (float)(r*normal[i][j][1] + p[1]);
			vertex[i][j][2] = (float)(r*normal[i][j][2] + p[2]);
		}
	}

	for(i=0; i<n-1; i++){
		for(int j=0; j<n-1; j++){
			glBegin(GL_POLYGON);

				glNormal3fv(normal[i][j]);
				glVertex3fv(vertex[i][j]);

				glNormal3fv(normal[i][j+1]);
				glVertex3fv(vertex[i][j+1]);

				glNormal3fv(normal[i+1][j+1]);
				glVertex3fv(vertex[i+1][j+1]);

				glNormal3fv(normal[i+1][j]);
				glVertex3fv(vertex[i+1][j]);

			glEnd();
		}
	}

	for(i=0; i<n; i++){
		delete[] vertex[i];
		delete[] normal[i];
	}
	delete[] vertex;
	delete[] normal;
}