// Smoother.cpp: implementation of the Smoother class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "rangeimage.h"
#include "Smoother.h"
#include "Normal.h"
#include "math.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Smoother::Smoother()
{

}

Smoother::~Smoother()
{

}

void Smoother::StMarc(CImage *in, int iter, float param, float thresh)
{
  int i;
  float Km, Kg, Ixx, Ixy, Iyy, Ix, Iy;
  int ROW, COL, m;

  in->getImageInfo(ROW, COL, m);

  long npix = (long )COL*ROW;

  float (** N)[3] = (float (**)[3])new float*[ROW];
   for( i=0; i<ROW; i++){
		N[i] = new float[COL][3];		
	}

  float (** n)[3] = (float (**)[3])new float*[ROW];
   for( i=0; i<ROW; i++){
		n[i] = new float[COL][3];		
	}
     

  float **C = new float*[ROW];
   for(i=0; i<ROW; i++){
		C[i] = new float[COL];		
	}
   
  float **Kmax = new float*[ROW];
   for(i=0; i<ROW; i++){
		Kmax[i] = new float[COL];		
	}

  float **Kmin = new float*[ROW];
   for(i=0; i<ROW; i++){
		Kmin[i] = new float[COL];		
	}

  int x, y;
  
  for ( y=0; y<ROW; y++){
    N[y][0][0] = 0;
    N[y][0][1] = 0;
    N[y][0][2] = -1;
	C[y][0] = 0;
  }

  for ( x=0; x<COL; x++)
  {	
    N[0][x][0] = 0;
	N[0][x][1] = 0;  
    N[0][x][2] = -1;
	C[0][x] = 0;
  }


   for ( y=1; y<ROW-1; y++)
     for ( x=1; x< COL-1; x++)
	  {			
		Ix = 0.5f*(in->getPixelVal(y, x+1) - in->getPixelVal(y, x-1));
		Iy = 0.5f*(in->getPixelVal(y+1, x) - in->getPixelVal(y-1, x));

		float r = 1.0 + Ix*Ix + Iy*Iy;
		double sr = sqrt(r);

		N[y][x][0] = (float)(Ix/sr);
		N[y][x][1] = (float)(Iy/sr);
		N[y][x][2] = (float)(-1.0/sr);        

        float Ixx = -0.5* (N[y][x+1][0]/N[y][x+1][2] - N[y][x-1][0]/N[y][x-1][2] );     
		float Iyy = -0.5* (N[y+1][x][1]/N[y+1][x][2] - N[y-1][x][1]/N[y-1][x][2] );
		float Ixy = -0.25* (N[y+1][x][0]/N[y+1][x][2] - N[y-1][x][0]/N[y-1][x][2] 
			               + N[y][x+1][1]/N[y][x+1][2]- N[y][x-1][1]/N[y][x-1][2]);
			
		float Kg = (Ixx*Iyy - Ixy*Ixy)/(r*r);
		float Km = ((1.0 + Iy * Iy) * Ixx + (1.0 + Ix * Ix) * Iyy - 
			         2.0 * Ix * Iy * Ixy) / (2.0 * r * sr);
 
		float Delta = 2.0* Km * Km - Kg;  /* half-sum of the squared principal curvatures */

		C[y][x] = exp(- param * Delta );
	  }
       
    for ( y = 0; y < ROW; y++)  
		for ( x = 0; x < COL; x++)
		{
	      n[y][x][2] = -1;
		}

  for (int it = 0; it < iter; it++)
      {
	  
 
  for (y = 1; y < ROW - 1; y++)
    for (int x = 1; x < COL - 1; x++)	
	  {
        float S=0;
        n[y][x][0] = 0;
        n[y][x][1] = 0;
        n[y][x][2] = 0;
	      
	for (i=-1; i <= 1; i++)	      
		for (int j=-1; j <= 1; j++) 
		  {
	          float dist = 1;
	     	  n[y][x][0] += N[y+j][x+i][0] * C[y+j][x+i] / dist;
              n[y][x][1] += N[y+j][x+i][1] * C[y+j][x+i] / dist;
              n[y][x][2] += N[y+j][x+i][2] * C[y+j][x+i] / dist;
              S += C[y+j][x+i] / dist;
	       }

            n[y][x][0] /= S;
            n[y][x][1] /= S;
            n[y][x][2] /= S;
	  }


	  for (y = 1; y <ROW - 1; y++) 
    	for ( x = 1; x < COL - 1; x++)                
          {
            Ix = - n[y][x][0] /n[y][x][2];
	        Iy = - n[y][x][1] /n[y][x][2];

	        float r  = 1 + Ix * Ix + Iy * Iy;
	        double sr = sqrt(r);

	       Ixx = - 0.5 * (n[y][x+1][0]/n[y][x+1][2] - n[y][x-1][0]/n[y][x-1][2]);
	       Iyy = - 0.5 * (n[y+1][x][1]/n[y+1][x][2] - n[y-1][x][1]/n[y-1][x][2]);
	       Ixy = - 0.25 * (n[y+1][x][0]/n[y+1][x][2] - n[y-1][x][0]/n[y-1][x][2] +
			               n[y][x+1][1]/n[y][x+1][2] - n[y][x-1][1]/n[y][x-1][2]);


	       Kg  = (Ixx * Iyy - Ixy * Ixy) / (r * r);
	       Km  = ((1 + Iy * Iy) * Ixx + (1 + Ix * Ix) * Iyy - 2 * Ix * Iy * Ixy) / (2. * r * sr);

           float Delta = 2.* Km * Km - Kg;	      
		   C[y][x] = exp(- param * Delta);

           N[y][x][0] = n[y][x][0];
           N[y][x][1] = n[y][x][1];
           N[y][x][2] = n[y][x][2];
          }
       }
 
  for (y=0;y < ROW; y++) 
	  delete [] n[y];
  delete [] n;

  float (** Vmax)[2] = (float (**)[2])new float*[ROW];
  for( i=0; i<ROW; i++)
		Vmax[i] = new float[COL][2];		
	

  float (** Vmin)[2] = (float (**)[2])new float*[ROW];
  for( i=0; i<ROW; i++)
		Vmin[i] = new float[COL][2];		

  for (y = 1; y < ROW - 1; y++)
    for ( x = 1; x < COL - 1; x++)    
	 {
        Ix = - N[y][x][0] /N[y][x][2];
	    Iy = - N[y][x][1] /N[y][x][2];

	    float r  = 1 + Ix * Ix + Iy * Iy;
	    double sr = sqrt(r);


	    Ixx = - 0.5 * (N[y][x+1][0]/N[y][x+1][2] - N[y][x-1][0]/N[y][x-1][2]);
	    Iyy = - 0.5 * (N[y+1][x][1]/N[y+1][x][2] - N[y-1][x][1]/N[y-1][x][2]);
	    Ixy = - 0.25 * (N[y+1][x][0]/N[y+1][x][2] - N[y-1][x][0]/N[y-1][x][2] +
			               N[y][x+1][1]/N[y][x+1][2] - N[y][x-1][1]/N[y][x-1][2]);


	     Kg  = (Ixx * Iyy - Ixy * Ixy) / (r * r);
	     Km  = ((1 + Iy * Iy) * Ixx + (1 + Ix * Ix) * Iyy - 2 * Ix * Iy * Ixy) / (2. * r * sr);

 	     float D = Km * Km - Kg;
	     if (D < 0.0) D = 0.;

	     Kmax[y][x] = Km + sqrt(D);
	     Kmin[y][x] = Km - sqrt(D);

	     float Amax   = Ixx / sr - (1+ Ix * Ix) * Kmax[y][x];
	     float Bmax   = Ixy / sr - Ix * Iy * Kmax[y][x];
	     float Cmax   = Iyy / sr - (1  + Iy * Iy) * Kmax[y][x];

	     if (fabs(Amax) > fabs(Cmax)){
		   Vmax[y][x][0] = -Bmax/(float)sqrt(1+Ix*Ix);
		   Vmax[y][x][1] =  Amax/(float)sqrt(1+Iy*Iy);
	     }
	     else{
		   Vmax[y][x][0] =  Cmax/(float)sqrt(1+Ix*Ix);
		   Vmax[y][x][1] = -Bmax/(float)sqrt(1+Iy*Iy);
	     }

	     float Amin   = Ixx / sr - (1+ Ix * Ix) * Kmin[y][x];
	     float Bmin   = Ixy / sr - Ix * Iy * Kmin[y][x];
	     float Cmin   = Iyy / sr - (1  + Iy * Iy) * Kmin[y][x];

	     if (fabs(Amin) > fabs(Cmin)){
		  Vmin[y][x][0] = -Bmin;
		  Vmin[y][x][1] =  Amin;
	     }
	     else{
		  Vmin[y][x][0] =  Cmin;
		  Vmin[y][x][1] = -Bmin;
	     }
	  }

  
	 for (y=0;y < ROW; y++){ 
	  delete [] N[y];
	  delete [] C[y];
	 }

	 delete [] C;
	 delete [] N;

	 /***** to extract ravines *****/
    non_max_suppr(in,Kmax,Vmax, thresh);
//    hysteresis(Rmax,Kmax);
    /******************************/

    /***** to extract ridges  *****/
    non_min_suppr(in,Kmin,Vmin, thresh);
//    hysteresis(R,Kmin);
    /******************************/
  
	 for (y=0;y < ROW; y++){ 
	  delete [] Vmax[y];
	  delete [] Vmin[y];
	  delete [] Kmin[y];		
	  delete [] Kmax[y];
	 }

	 delete [] Kmax;
	 delete [] Kmin;
	 delete [] Vmax;
	 delete [] Vmin;

}

 float eps = 0.00000001;
 float THRESH = 0.000001;

void Smoother::non_max_suppr(CImage *im, float **Kmax, float (**Vmax)[2], float T)
{
  int ROW,COL,x,y, tmp;
  float tg, lappr1, lappr2;

 
  im->getImageInfo(ROW, COL, tmp);

  for (y = 1; y < ROW - 1; y++)
    for (x = 1; x < COL - 1; x++)
    {	
	  if (fabs(Vmax[y][x][0]) < eps) {
	      lappr1 = Kmax[y+1][x]; lappr2 = Kmax[y-1][x];
	  }
	  else {
	      tg = Vmax[y][x][1] / Vmax[y][x][0];

	      if ((tg >= 0.) && (tg < 1.)) {
	         lappr1 = Kmax[y+1][x+1] * (1.- tg) + tg * Kmax[y][x+1];
	         lappr2 = Kmax[y-1][x-1] * (1.- tg) + tg * Kmax[y][x-1];
		  }
          else if ( tg >= 1.) {
	         lappr1 = (Kmax[y+1][x+1] * (tg - 1.) + Kmax[y+1][x]) / tg;
	         lappr2 = (Kmax[y-1][x-1] * (tg - 1.) + Kmax[y-1][x]) / tg;

		  }
          else if ((tg <= 0.) && (tg >  -1.)) {
	         lappr1 = Kmax[y+1][x-1] * (1.+ tg) - tg * Kmax[y][x-1];
	         lappr2 = Kmax[y-1][x+1] * (1.+ tg) - tg * Kmax[y][x+1];

		  }
	     else if ( tg <= -1.) {
	         lappr1 = (Kmax[y+1][x-1] * (tg + 1.) - Kmax[y+1][x]) / tg;
	         lappr2 = (Kmax[y-1][x+1] * (tg + 1.) - Kmax[y-1][x]) / tg;
		 }
	  }

	if ( Kmax[y][x] > (lappr1+ THRESH) &&
		 Kmax[y][x] > (lappr2+ THRESH) &&
		(Kmax[y][x] > T))

	     im->setFeature(y, x, RIDGE);
      }
}

void Smoother::non_min_suppr(CImage *im, float **Kmin, float (**Vmin )[2], float T)
{
  int ROW, COL, x,y, tmp;
  float tg, lappr1, lappr2;

 
  im->getImageInfo(ROW, COL, tmp);

  for (y = 1; y < ROW - 1; y++)
    for (x = 1; x < COL - 1; x++)
    {	
	  if (fabs(Vmin[y][x][0]) < eps) {
	      lappr1 = Kmin[y+1][x]; lappr2 = Kmin[y-1][x];
	  }
	  else {
	      tg = Vmin[y][x][1] / Vmin[y][x][0];

	      if ((tg >= 0.) && (tg < 1.)) {
	         lappr1 = Kmin[y+1][x+1] * (1.- tg) + tg * Kmin[y][x+1];
	         lappr2 = Kmin[y-1][x-1] * (1.- tg) + tg * Kmin[y][x-1];
		  }
          else if ( tg >= 1.) {
	         lappr1 = (Kmin[y+1][x+1] * (tg - 1.) + Kmin[y+1][x]) / tg;
	         lappr2 = (Kmin[y-1][x-1] * (tg - 1.) + Kmin[y-1][x]) / tg;

		  }
          else if ((tg <= 0.) && (tg >  -1.)) {
	         lappr1 = Kmin[y+1][x-1] * (1.+ tg) - tg * Kmin[y][x-1];
	         lappr2 = Kmin[y-1][x+1] * (1.+ tg) - tg * Kmin[y][x+1];

		  }
	     else if ( tg <= -1.) {
	         lappr1 = (Kmin[y+1][x-1] * (tg + 1.) - Kmin[y+1][x]) / tg;
	         lappr2 = (Kmin[y-1][x+1] * (tg + 1.) - Kmin[y-1][x]) / tg;
		 }
	  }

	if ( lappr1 > (Kmin[y][x]+ THRESH) &&
		 lappr2 > (Kmin[y][x]+ THRESH) &&
		(Kmin[y][x] < -T))

	     im->setFeature(y, x, RAVINE);
      }
}


